import base64
import datetime
import tempfile
from typing import List, Dict, Any, Optional

from imapclient.response_types import Envelope
from mailparser import mailparser

from email_miner.email import Email, Attachment


def decode_attachment_data(attachment) -> bytes:
    """
    Decode a raw attachment into usable bytes.

    :param attachment:
    :return: the bytes containing file data
    """
    if attachment['content_transfer_encoding'] == 'base64':
        return base64.b64decode(attachment['payload'])
    return attachment['payload'].encode('utf-8')


def empty_if_none(input_value=None) -> List[Any]:
    """
    Return an empty list of the input value is None.

    :param input_value:
    :return:
    """
    if input_value is not None:
        return input_value
    return []


def decode_bytes_str(input_value=None, charset='utf-8') -> Optional[str]:
    """
    Decode input value to string using the given charset
    :param charset: charset to use while decoding (defaults to utf-8)
    :param input_value:
    :return:
    """
    if input_value is not None and not isinstance(input_value, str):
        return input_value.decode(charset)
    return input_value


def parse_emails(raw_emails: Dict[int, Dict], in_memory: bool = True) -> List[Email]:
    """
    Parse raw emails into easy-to-use classes.

    :param raw_emails:
    :param in_memory:
    :return: list of Email classes.
    """
    emails = []

    for mail_id, raw_email in raw_emails.items():
        email = Email()
        envelope: Envelope = raw_email[b'ENVELOPE']
        internal_date: datetime.datetime = raw_email[b'INTERNALDATE']

        email.raw_envelope = envelope

        email.id = int(mail_id)
        email.date = envelope.date
        email.internal_date = internal_date
        email.subject = decode_bytes_str(envelope.subject)
        email.from_addresses = list(empty_if_none(envelope.from_))
        email.reply_to_addresses = list(empty_if_none(envelope.reply_to))
        email.to_addresses = list(empty_if_none(envelope.to))
        email.cc_addresses = list(empty_if_none(envelope.cc))
        email.bcc_addresses = list(empty_if_none(envelope.bcc))
        email.in_reply_to = decode_bytes_str(envelope.in_reply_to)
        email.uuid = decode_bytes_str(envelope.message_id)

        if b'BODY[]' in raw_email:
            email.raw_body = raw_email[b'BODY[]']
            parsed = mailparser.parse_from_bytes(email.raw_body)
            for raw_attachment in parsed.attachments:
                attachment = Attachment()
                attachment.file_name = raw_attachment['filename']
                attachment.content_type = raw_attachment['mail_content_type']
                data_bytes = decode_attachment_data(raw_attachment)
                if in_memory:
                    attachment.file_content = data_bytes
                else:
                    fd, path = tempfile.mkstemp()
                    fd.write(data_bytes)
                    attachment.file_path = path
                    fd.close()
                email.attachments.append(attachment)

        emails.append(email)

    return emails



import contextlib
import imaplib
import logging
import ssl
from typing import List

from imapclient import IMAPClient, SEEN

from email_miner.email import Email
from email_miner.parse import parse_emails


class Miner:
    """create a new email miner instance that can be used to traverse mails"""

    imap: IMAPClient = None

    def __init__(self, hostname: str,
                 username: str,
                 password: str,
                 port: int = imaplib.IMAP4_SSL_PORT,
                 use_ssl: bool = True,
                 verify: bool = True,
                 log_level: int = None):
        """
        Create a new instance of the miner.

        :param hostname: the hostname of the imap server to connect to
        :param username: the user to login as
        :param password:
        :param port: the port to connect to. (defaults to 993)
        :param use_ssl: whether to use SSL to connect (defaults to True)
        :param verify: whether to verify the SSL certificates (defaults to False)
        """

        if log_level is not None:
            logging.basicConfig(
                format='%(asctime)s - %(levelname)s: %(message)s',
                level=log_level
            )

        ssl_context = ssl.create_default_context()

        if not verify:
            # disable hostname check. certificate may not match hostname.
            ssl_context.check_hostname = False
            # disable certificate authority verification. certificate maybe issued by unknown CA
            ssl_context.verify_mode = ssl.CERT_NONE

        self.imap = IMAPClient(host=hostname, port=port, ssl=use_ssl, ssl_context=ssl_context)
        self.imap.login(username, password)

    @contextlib.contextmanager
    def folder(self, folder_name: str, read_only: bool = True):
        """
        Switch to a specific folder.

        :param folder_name: name of the folder to switch to
        :param read_only: read-only mode will not mark emails as read even after retrieval
        :return:
        """
        try:
            yield self.imap.select_folder(folder_name, read_only)
        finally:
            self.imap.close_folder()

    @contextlib.contextmanager
    def inbox(self, read_only: bool = True):
        """
        Switch to the inbox folder.

        :param read_only: read-only mode will not mark emails as read even after retrieval
        :return:
        """
        try:
            yield self.imap.select_folder('inbox', read_only)
        finally:
            self.imap.close_folder()

    def mark_as_unread(self, message_ids: List[int]):
        """
        Mark the given message IDs as unread by removing the SEEN flag.

        :param message_ids:
        :return:
        """
        self.imap.remove_flags(message_ids, [SEEN])

    def mark_as_read(self, message_ids: List[int]):
        """
        Mark the given message IDs as read by adding the SEEN flag.

        :param message_ids:
        :return:
        """
        self.imap.add_flags(message_ids, [SEEN])

    def get_emails(self, unread_only: bool = True,
                   with_body: bool = False,
                   keep_as_unread: bool = False,
                   in_memory: bool = True) -> List[Email]:
        """
        Get emails from the selected folder.

        :param keep_as_unread: keep any retrieved emails as unread in the mailbox.
        :param unread_only: choose only to retrieve unread mails
        :param with_body: read-only mode will not mark emails as read even after retrieval
        :param in_memory: store the parsed attachments in-memory as bytes or to a temp file locally
        :return:
        """
        ids = self.imap.search('(UNSEEN)' if unread_only else 'ALL')
        flags = ['ENVELOPE', 'FLAGS', 'UID', 'INTERNALDATE']

        if with_body:
            flags.append('BODY[]')

        response = self.imap.fetch(ids, flags)

        try:
            if keep_as_unread:
                self.mark_as_unread(ids)
            else:
                self.mark_as_read(ids)
        except Exception:
            # will throw an exception if folder in read-only mode. so ignore.
            pass

        return parse_emails(response, in_memory)

    def __enter__(self):
        """
        return the instance of the miner for use as a context manager.

        :return:
        """
        return self

    def __exit__(self, *args):
        """
        Close folder and logout on exit when used as a context manager.

        :param args:
        :return:
        """
        if self.imap is not None:
            try:
                self.imap.close_folder()
            except:
                pass
            self.imap.logout()



import datetime
from typing import List

from imapclient.response_types import Address, Envelope


class Attachment:
    """
    :ivar file_name: Name of the attachment file
    :ivar content_type: content type (mime) of the attachment file
    :ivar file_content: content of the file as bytes
    :ivar file_path: file path of the locally stored copy of the attachment
    """
    file_name: str
    content_type: str
    file_content: bytes
    file_path: str = None


class Email:
    """
    :ivar id: Unique ID referring to the mail in the mailbox
    :ivar date: A datetime instance that represents the "Date" header.
    :ivar subject: A string that contains the "Subject" header.
    :ivar from: A tuple of Address objects that represent one or more
      addresses from the "From" header, or None if header does not exist.
    :ivar reply_to: As for from but represents the "Reply-To" header.
    :ivar to: As for from but represents the "To" header.
    :ivar cc: As for from but represents the "Cc" header.
    :ivar bcc: As for from but represents the "Bcc" recipients.
    :ivar in_reply_to: A string that contains the "In-Reply-To" header.
    :ivar message_id: A string that contains the "Message-Id" header.
    """
    id: int
    date: datetime.datetime
    internal_date: datetime.datetime
    subject: str
    from_addresses: List[Address] = []
    reply_to_addresses: List[Address] = []
    to_addresses: List[Address] = []
    cc_addresses: List[Address] = []
    bcc_addresses: List[Address] = []
    in_reply_to: str
    message_id: str

    raw_envelope: Envelope
    raw_body: bytes

    attachments: List[Attachment] = []
